(***********************************************************************)
(*                                                                     *)
(*                          Sklml                                      *)
(*                                                                     *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*             Quentin Carbonneaux (Quentin.Carbonneaux@inria.fr)      *)
(*                                                                     *)
(* This program is free software.                                      *)
(* (see the file LICENSE in the root directory of the distribution).   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: skl.mli,v 1.6 2010-10-22 07:07:36 weis Exp $ *)

(** {3 The Sklml skeleton algebra} *)

(** This is the signature of the skeleton algebra, the set of tools offered
   to the Sklml programmer to define the parallel behaviour of programs. *)

(** {6 Basic skeletons} *)

type ('a, 'b) skel = private 'a Stream.t -> 'b Stream.t;;
(** A skeleton from A to B is a lazy stream processing function from
    lazy stream of A to lazy stream of B. *)

val make : ('a Stream.t -> 'b Stream.t) -> ('a, 'b) skel;;
val from : ('a, 'b) skel -> ('a Stream.t -> 'b Stream.t);;

val lift : ('a -> 'b) -> ('a, 'b) skel;;
(** Lift a regular function from A to B to a skeleton from A to B.

Streams.map as
Streams.lift : ('a -> 'b) -> ('a stream -> 'b stream);;
    [lift f] is equivalent to [make (Streams.lift f)].
*)

val unlift : ('a, 'b) skel -> ('a -> 'b);;
(** Unlift a skeleton from stream of A to stream of B to a
    regular function from A to B.
    Presumably, for each [x] in A, [unlift fskl] applies [fskl] to the
    singleton stream that contains only [x] and returns the unique
    element of the resulting stream.

Streams.unlift : ('a stream -> 'b stream) -> ('a -> 'b)
[unlift skl] is equivalent to [Streams.unlift (from skl)].

*)

val ( ||| ) : ('a, 'b) skel -> ('b, 'c) skel -> ('a, 'c) skel;;
(** [skl1 ||| skl2] composes the two skeletons [skl1] and [skl2]:
  it applies [skl1] then [skl2] to the [Streams.t] argument.

[( ||| )] provides program parallelism, since the two skeletons run in
parallel. *)

val farm : ('a, 'b) skel * int -> ('a, 'b) skel;;
(** Given a skeleton [skl] and a positive integer [n],
   [farm (skl, n)] computes [skl] on each element of the input stream in
   parallel with [n] processes. If the input stream is [x1; x2; x3; ...; xk],
   the output stream is [f x1; f x2; f x3; ...; f xk] where [f = unlift skl].

   [farm] provides program parallelism: it simultaneously applies [n]
   instances of the skeleton [skl] to the elements of the input stream.
   Rq: \forall skl, n > 0, s, farm (skl, n) s = skl s.
   Hence, a minimal specification would be:
   [let farm (skl, n) = skl;;]
   which is a semantically correct implementation for farm. *)

val farm_vector : ('a, 'b) skel * int -> ('a array, 'b array) skel;;
(** Given a skeleton [skl] and a positive integer [n],
  [farm_vector (skl, n)] maps [skl] on each vector of the input stream in parallel:
  it computes [skl] on each element of the input vector and returns the
  corresponding vector result as the next element of the output stream.
  [farm_vector] combines data and program parallelism. *)

(*
val Array.fold_left as
    fold_left_vect :        ('a * 'b -> 'a) -> 'a -> 'b vect -> 'a
;;
val Streams.fold_left as
    fold_vect_lifted :      ('a * 'b -> 'a) -> 'a -> ('b vect) stream -> 'a stream
;;

val fold_left_vect_skel_lifted : ('a * 'b -> 'a) -> 'a -> ('b vect, 'a) skel
(** [fold_left_vect_skel_lifted f] is equivalent to [fold_left_vect_skel (f, 1)]. *)
;;

val Skl.fold_left_vect as
    fold_left_vect_skel :   ('a * 'b -> 'a, int) -> 'a -> ('b vect, 'a) skel
;;
*)

val reduce_vector : ('a * 'b, 'a) skel * int -> 'a -> ('b array, 'a) skel;;
(* val reduce_vector :
    ('a * 'a, 'a) skel * int -> 'a -> ('a array, 'a) skel;; *)
(** [reduce_vector (f, n) z sv] computes for each element [v] of [sv], the
  value [f z (f v.(0) (f v.(1) (f ... )))].
  [f] must be a binary associative commutative operator, if [v] is empty
  [reduce_vector] outputs [z]. *)

val loop : ('a -> bool) * ('a, 'a) skel -> ('a, 'a) skel;;
(** Given a predicate function [p] and a skeleton [skl],
    [loop (p, skl)] returns the stream processing function that loops on each
    element of the input stream until predicate [p] holds.
    In short, for each element [x] of the input stream, the skeleton
    [loop (p, skl)] computes [f^n (x)] for the least integer [n] strictly
    greater than 1, such that [p (f^n (x))] holds, where [f = unlift skl]. *)

val parfun : ('a, 'b) skel -> 'a Stream.t -> 'b Stream.t;;
(** Turns a skeleton into a regular function from [Streams.t] to [Streams.t]. *)

val pardo : (unit -> unit) -> unit;;
(** [pardo] just applies its function argument to launch the computational
  network. *)

(** {6 Skeletons with initialisation} *)

(** A skeleton with initialisation is a function that returns a skeleton each
  time it is applied to the value [()]. If [skli] is a skeleton with
  initialisation, then [skli ()] is simply a skeleton which may encapsulate
  some arbitrary initial computation. Alternatively, [skli] can be considered
  as a skeleton generator, i.e. a function that generates a (presumably new)
  skeleton each time it is called. In the parallel interpretation, [skli] may
  presumably generate new worker nodes in the computational network. *)

type ('a, 'b) skel_init = private unit -> ('a, 'b) skel;;

(** A [skel_init] value is a {e skeleton with initialisation} or {e skeleton
  generator}, i.e. a function that generates a (presumably new) skeleton each
  time it receives its first parameter. Put it another way: a [skel_init]
  value is a skeleton that needs some initialisation that takes place
  when the [()] argument is applied; then a properly initialized skeleton is
  returned and can be used normally. *)

val lift_init : (unit -> 'a -> 'b) -> ('a, 'b) skel_init;;
(** The lift operator for a function with initialisation:
  given a function [f] with initialisation from A to B, [seq_init f] is a
  skeleton  with initialisation from streams of A to streams of B.
[lift_init fi = fun () -> make (Streams.lift (fi ()))]
 *)

val unlift_init : ('a, 'b) skel_init -> (unit -> 'a -> 'b);;
(** [unlift_init skli =
      fun () -> Streams.unlift (from (skli ()))] *)

val farm_init : ('a, 'b) skel_init * int -> ('a, 'b) skel;;
(** The [farm] corresponding skeleton for skeleton generators.
  The skeleton generator argument is iteratively applied to generate
  all the workers necessary for the [farm_init] computational node.
*)

val farm_vector_init : ('a, 'b) skel_init * int -> ('a array, 'b array) skel;;
(** Same as [farm_vector] for skeleton generators: the skeleton generator
  argument is iteratively applied (to the value "()") to generate
  all the workers necessary for the [farm_vector] computational node. *)

val reduce_vector_init : ('a * 'a, 'a) skel_init * int -> 'a -> ('a array, 'a) skel;;
(** The [reduce_vector] corresponding skeleton for skeleton generators.
  The skeleton generator is iteratively applied to generate all the workers
  necessary for the [reduce_vector_init] computational node.
*)

(** {6 Obsolete skeletons} *)

(* Those skeletons are aliases for skeleton above.
   Provided for backward compatibility.
   Do not use in new programs. *)

val seq : ('a -> 'b) -> ('a, 'b) skel;;
(** An obsolete alias for [Skl.lift]. *)

val mapvector : ('a, 'b) skel * int -> ('a array, 'b array) skel;;
(** An obsolete alias for [Skl.farm_vector]. *)
val reducevector : ('a * 'a, 'a) skel * int -> 'a -> ('a array, 'a) skel;;
(** An obsolete alias for [Skl.reduce_vector]. *)
